﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace Romfarer {
    /// <summary>
    /// Control and configuration for all attached lazors and modules.
    /// </summary>
    class Config {

        private LazorSystem system;
		public List<CFG> cfgActive = new List<CFG>(); //CFG's that are running

		#region lazor fields
		private List<LazorAimable> lazors = new List<LazorAimable>();
        public Dictionary<int, LazorAimable> lazorsById = new Dictionary<int, LazorAimable>(); // key=LazorAimable.GetInstanceID()
        private Dictionary<int, List<LazorAimable>> lazorCounterparts = new Dictionary<int, List<LazorAimable>>(); //all lazors on vessel
        public Dictionary<int, List<LazorAimable>> LazorCounterparts {
            get { return lazorCounterparts; }
        }
        public HashSet<int> claimedAndActiveLazorGroups = new HashSet<int>();
        public Dictionary<int, GameObject> identifyEffects = new Dictionary<int, GameObject>();
        private List<CFG> cfgPrefabs = new List<CFG>(); //all possible CFG's
        public HashSet<int> unassignedLazorGroups = new HashSet<int>();
        public HashSet<int> focusedListItems = new HashSet<int>(); //list of LazorAimable.GetInstanceID()
        public HashSet<int> focusedListGroups = new HashSet<int>();
        public HashSet<int> expandedListGroups = new HashSet<int>();
		#endregion

		#region modules fields
        public List<ModuleInfo> modulesT2Active = new List<ModuleInfo>();
        public List<ModuleInfo> modulesT1Active = new List<ModuleInfo>();
        public Dictionary<int, List<Part>> modulesCounterparts = new Dictionary<int, List<Part>>();
		public Dictionary<Part, PidControllers> pids = new Dictionary<Part, PidControllers>();
        internal class ModuleInfo {
            public string name; //internal unique name
            public string title; //display name
            public bool toggle; //on/off
            public List<int> counterparts; //referenced in moduleCounterparts
            public List<string> behaviours; //partmodules
            public bool proximitydetonate; //tmp
            public float proximitydetonate_range = 0;//tmp
            public float proximitydetonate_maxrange = 0;//tmp
            public bool triggerOn;
            public string triggerName;
            public ModuleInfo(string name, string title, bool toggle, List<int> counterparts, List<string> behaviours) {
                this.name = name;
                this.title = title;
                this.toggle = toggle;
                this.counterparts = counterparts;
                this.behaviours = behaviours;
            }
        }
        internal class PidControllers {
			public PIDclamp yaw = new PIDclamp("Missile Yaw", 0.2f, 0.02f, 0.6f, 1f);
			public PIDclamp pitch = new PIDclamp("Missile Pitch", 0.2f, 0.02f, 0.6f, 1f);
            public Vector3 acc = Vector3.zero;
            public PidControllers() { }
        }
		#endregion


        public Config(LazorSystem system) {
			this.system = system;

			foreach(Part p in system.vessel.parts) {
				if(p is LazorAimable) {
					LazorAimable q = p as LazorAimable;
					lazors.Add(q);
					q.Initialize(system);
				}
			}

            Register(lazors); //add all lazors to lazorCounterparts
            //Set up the CFG's
            cfgPrefabs.Add(new CFG_blue(system));
            cfgPrefabs.Add(new CFG_green(system));
            cfgPrefabs.Add(new CFG_red(system));
            cfgPrefabs.Add(new CFG_magenta(system));
            cfgPrefabs.Add(new CFG_yellow(system));
            foreach (CFG c in cfgPrefabs)
                AddCFG((CFG)c.GetType().GetConstructors()[0].Invoke(new[] { system }));

            unassignedLazorGroups.UnionWith(lazorCounterparts.Keys); //add all lazor groups to unassigned set
            foreach (CFG cfg in cfgActive) //remove active lazor groups from unassigned set
                unassignedLazorGroups.RemoveWhere(x => cfg.GetClaimedLazors().Contains(x));

            //Modules
            int count = RegisterController(0, "LazorControlT2", modulesT2Active, true);
            RegisterBehaviour("LazorWarhead", modulesT2Active);
            RegisterController(count, "LazorControlT1", modulesT1Active, true);
            RegisterBehaviour("LazorTriggerContinous", modulesT1Active);

            //add docking cam to all docking ports
            /*foreach (Part p in system.vessel.parts) {
                if (p.Modules.Contains("ModuleDockingNode")) {
                    if (!p.Modules.Contains("LazorSystemDockingCamera")) {
                        MonoBehaviour.print("[Adding docking camera to " + p.name + "]");
                        PartModule pm = p.AddModule("LazorSystemDockingCamera");
                    }
                }
            }*/
        }

		
		#region lazors
		
		public void ResetLazorOrientations(){
			foreach (LazorAimable l in lazors)
				l.Lazor.ResetOrientation();
		}

		//shut down all lazors in all configs handled by this cfg
		public void ShutDown() {
			foreach(CFG cfg in cfgActive) {
				foreach(int i in cfg.GetClaimedLazors())
					foreach(LazorAimable l in LazorCounterparts[i])
						l.Lazor.Off();
			}
		}

		public void ActivateCFG(CFG cfg) {
			claimedAndActiveLazorGroups.UnionWith(cfg.GetClaimedLazors());
		}

		public void DeactivateCFG(CFG cfg) {
			claimedAndActiveLazorGroups.RemoveWhere(l => cfg.GetClaimedLazors().Contains(l)); //remove all of c's claimed lazors from lazorlist
		}
		
        public bool AddCFG(CFG cfg) {
            if (cfgActive.Exists(x => x.name == cfg.name)) {
                CFG existing = cfgActive.Find(x => x.name == cfg.name);
                if (!existing.Detect(LazorCounterparts)) { //if cfg was enabled on docked vessel, but no lazors of type present after dock
                    cfgActive.RemoveAll(x => x.name == cfg.name);
                    return false;
                } else { //disables on dock, but also on load =/
                    existing.Reset();
					if(existing.Enabled)
						claimedAndActiveLazorGroups.UnionWith(existing.GetClaimedLazors());
                }
                return true;
            } else {
                if (cfg.Detect(LazorCounterparts)) {
                    cfgActive.Add(cfg);
                    return true;
                }
            }
            return false;
        }

		public bool TryAddCFG(int i) { //★ or ▲ button...called when pressing the up button in unassigned lazors, and there are no existing cfg
            foreach (CFG c in cfgPrefabs) {
                if (c.Claimable(LazorCounterparts[i][0].Lazor.GetColor())) {
                    CFG newCFG = (CFG)c.GetType().GetConstructors()[0].Invoke(new[] { system });
                    if (newCFG.Detect(new Dictionary<int, List<LazorAimable>>() { { i, LazorCounterparts[i] } })) {
                        cfgActive.Add(newCFG);
                        unassignedLazorGroups.Remove(i);
                        return true;
                    }
                }
            }
            return false;
        }

		public bool TryAddToExisting(int i) { //▲ button...attempts to add the group (i) to an existing cfg (have already fount out it can only be added to ONE)
            foreach (CFG cfg in cfgActive) {
                if (cfg.Claimable(lazorCounterparts[i][0].Lazor.GetColor())) {
                    if (cfg.ClaimLazors(i)) {
						if(cfg.Enabled)
							claimedAndActiveLazorGroups.Add(i);
                        unassignedLazorGroups.Remove(i);
                        return true;
                    }
                }
            }
            return false;
        }

		public bool TryAddToExistingSpecific(System.Object[] o) { //∈ button ... attempts to add the group (i) to an existing cfg (have already fount out it can only be added to ONE)
            int i = (int)o[0];
            CFG cfg = (CFG)o[1];
            if (cfg.Claimable(lazorCounterparts[i][0].Lazor.GetColor())) {
                if (cfg.ClaimLazors(i)) {
					if(cfg.Enabled)
						claimedAndActiveLazorGroups.Add(i);
                    unassignedLazorGroups.Remove(i);
                    return true;
                }
            }
            return false;
        }

		public bool TryRemove(int i) { //▼ button pressed...remove group (i) from cfg
            foreach (CFG cfg in cfgActive) {
                if (cfg.UnclaimLazors(i)) {
                    unassignedLazorGroups.Add(i);
                    if (cfg.GetClaimedLazors().Count == 0) {
                        cfg.Dispose();
                        cfgActive.Remove(cfg);
						claimedAndActiveLazorGroups.Remove(i);
                    }
                    return true;
                }
            }
            return false;
        }

		//Add to lazorCounterparts and lazorsById
        public void Register(List<LazorAimable> lazors) {
            List<LazorAimable> tmp = new List<LazorAimable>();
            lazorsById.Clear(); //hmm not necessary?
			lazorCounterparts.Clear();
            int counter = 0;
            foreach (LazorAimable l in lazors) {
                if (!tmp.Contains(l)) {
                    lazorCounterparts.Add(++counter, new List<Part>(l.symmetryCounterparts).ConvertAll<LazorAimable>(o => (LazorAimable)o));
                    lazorCounterparts[counter].Add(l);
                }
                tmp.AddRange(new List<Part>(l.symmetryCounterparts).ConvertAll<LazorAimable>(o => (LazorAimable)o));
                tmp.Add(l);
                lazorsById.Add(l.GetInstanceID(), l);
                AddIdentifyEffect(l);
            }
        }

		//Remove from lazorCounterparts and lazorsById
		private List<int> removesetUnRegister = new List<int>();
		public void UnRegister(LazorAimable l) {
			lazorsById.Remove(l.GetInstanceID());
			focusedListItems.Remove(l.GetInstanceID());
			removesetUnRegister.Clear();
			foreach(int i in lazorCounterparts.Keys) {
				lazorCounterparts[i].Remove(l);
				if(lazorCounterparts[i].Count == 0)
					removesetUnRegister.Add(i);
			}
			foreach(int x in removesetUnRegister){
				unassignedLazorGroups.Remove(x);
				focusedListGroups.Remove(x);
				expandedListGroups.Remove(x);
				lazorCounterparts.Remove(x);
				CFG stuff = cfgActive.Find(c => c.GetClaimedLazors().Contains(x));
				if(stuff != null)
					stuff.UnclaimLazors(x);

			}
			foreach(CFG cfg in cfgActive) {
				cfg.RemoveLazorRef(l);
			}
		}

		/*public void LazorDisconnected(LazorAimable l) {
			lazors.Remove(l);
			UnRegister(l);
		}*/

       	//Graphically identifies the lazor
       	public void Identify(LazorAimable lazor) {
			if(lazor != null){
	           ParticleEmitter pe = system.config.identifyEffects[lazor.GetInstanceID()].particleEmitter;
	           pe.Emit(10);
			}
       	}

       	public void IdentifyContinuous(LazorAimable lazor) {
			if(lazor != null){
	           	ParticleEmitter pe = system.config.identifyEffects[lazor.GetInstanceID()].particleEmitter;
	           	pe.Emit(1);
			}
       	}

		private void AddIdentifyEffect(LazorAimable lazor) {
			GameObject o = Part.Instantiate(UnityEngine.Resources.Load("Effects/fx_exhaustFlame_blue")) as GameObject;
			o.transform.parent = lazor.transform;
			o.transform.localEulerAngles = Vector3.zero;
			o.transform.localPosition = new Vector3(0, 0.13f, 0);
			ParticleEmitter emitter = o.particleEmitter;
			emitter.emit = false;
			emitter.minSize = 0.13f;
			emitter.maxSize = 0.15f;
			emitter.minEnergy = 1.8F;
			emitter.maxEnergy = 2f;
			emitter.rndVelocity = Vector3.one * 0.02f;
			emitter.localVelocity = Vector3.zero;
			emitter.useWorldSpace = false;
			(o.GetComponent<ParticleAnimator>() as ParticleAnimator).sizeGrow = -0.6F;
			ParticleAnimator flare_emitterAnimator = emitter.GetComponent<ParticleAnimator>();
			Color[] emitterColors = flare_emitterAnimator.colorAnimation;
			emitterColors[0] = lazor.Lazor.GetColor();
			emitterColors[1] = lazor.Lazor.GetColor();
			emitterColors[2] = lazor.Lazor.GetColor();
			emitterColors[3] = lazor.Lazor.GetColor();
			flare_emitterAnimator.colorAnimation = emitterColors;
			identifyEffects.Add(lazor.GetInstanceID(), o);
		}

		#endregion
		
		#region events

		//Reconfigures by taking lazors away
		public void Reconfigure() {
			RuntimeLogic.Instance.Log (system + " on " + system.vessel.vesselName + " : Config : Reconfiguring", true);
			List<CFG> removeCFG = new List<CFG> ();
			for (int i = 0; i < cfgActive.Count; i++) { //iterate: active cfg
				int[] claimed = cfgActive[i].GetClaimedLazors().ToArray(); //claimed[c] = keys in lazorCounterparts
				for (int c = 0; c < claimed.Length; c++) { //iterate: claimed lazor groups
					//RuntimeLogic.Instance.Log ("check "+lazorCounterparts[claimed[c]].Count + " lazors", false);
					List<LazorAimable> removeLazors = new List<LazorAimable> ();
					for(int l = 0; l < lazorCounterparts[claimed[c]].Count; l++){
						LazorAimable lazor = lazorCounterparts [claimed[c]].ElementAt(l);
						//RuntimeLogic.Instance.Log ("check "+lazor.GetInstanceID()+" "+l, false);
						if (lazor == null)
							focusedListItems.Remove (lazor.GetInstanceID());
						else if (lazor.vessel != system.vessel){ //lazor.localRoot != system.localRoot) {
							//RuntimeLogic.Instance.Log ("x "+lazor.GetInstanceID(), false);
							removeLazors.Add (lazor);
							focusedListItems.Remove(lazor.GetInstanceID());
							focusedListGroups.Remove(claimed[c]);
							expandedListGroups.Remove(claimed[c]);
							cfgActive[i].RemoveLazorRef(lazor);
						}
					}
					lazorCounterparts [claimed[c]].RemoveAll (x => removeLazors.Contains(x));
					if(lazorCounterparts [claimed[c]].Count == 0){
						//RuntimeLogic.Instance.Log ("y ", false);
						cfgActive[i].GetClaimedLazors().Remove(claimed[c]);
						claimedAndActiveLazorGroups.Remove(claimed[c]);
						unassignedLazorGroups.Remove(claimed[c]);
					}
				}
				if (cfgActive [i].GetClaimedLazors ().Count == 0)
					removeCFG.Add (cfgActive[i]);
			}
			cfgActive.RemoveAll (x => removeCFG.Contains (x));
		}



		public void OnFixedUpdate() {
			if (system.lazorOn) {
				//Identify Effects
				if (system.guiConfig.toggleIdentify) {
					if (focusedListItems.Count > 0) {
						foreach (int i in focusedListItems) {
							LazorAimable l = system.config.lazorsById[i];
							if (l != null)
								IdentifyContinuous(l);
						}
					} else if (focusedListGroups.Count > 0) {
						foreach (int i in focusedListGroups)
							foreach (LazorAimable l in system.config.lazorCounterparts[i])
								if (l != null)
									IdentifyContinuous(l);
					}
				}
				//Send control updates to all lazors in CFG's
				int activeLazorCount = 0;
				foreach (CFG c in cfgActive) {
					activeLazorCount += c.OnFixedUpdate();
				}
				system.RequestEnergyDelta(activeLazorCount * 0.0012d * (system.lazorRange / 10000 + 1)); //draw energy from system for lazors.
			}
		}

		public void OnTargetSwitch(){
			foreach (CFG cfg in cfgActive)
				cfg.OnTargetSwitch();
		}
		
		public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData){
			foreach (CFG cfg in cfgActive)
				cfg.onFlightStateLoad(parsedData);
		}
		
		public void onFlightStateSave(Dictionary<string, KSPParseable> parsedData){
			List<Type> seen = new List<Type>();
			foreach(CFG cfg in cfgActive) {
				if(!seen.Contains(cfg.GetType())) {
					cfg.onFlightStateSave(parsedData);
					seen.Add(cfg.GetType());
				}
			}
			seen.Clear();
		}
		
		public void DisposeCFGs() {
			foreach(CFG cfg in cfgActive)
				if(cfg is IDisposable)
					((IDisposable)cfg).Dispose();
		}
		
		public void DrawCFGGuis(){
			foreach (CFG cfg in cfgActive)
				cfg.DrawGUI();
		}
		#endregion
		

		#region modules

        public int RegisterController(int counter, string moduleName, List<ModuleInfo> list, bool enabled) {
            int c = counter;
            foreach (Part p in system.vessel.parts) {
                if (p.Modules.Contains(moduleName)) {
                    if (!DictionaryContains(modulesCounterparts.Values, p)){
                        modulesCounterparts.Add(++c, new List<Part>(p.symmetryCounterparts)); //add found part/conterparts to dictionary
                        modulesCounterparts[c].Add(p);
                        if (!list.Exists(x => x.name == p.partInfo.name))
                            list.Add(new ModuleInfo(p.partInfo.name, p.partInfo.title, enabled, new List<int>(new[] { c }), new List<string>(new[] { moduleName })));
                        else
                            (list.Find(x => x.name == p.partInfo.name).counterparts).Add(c);
                    }
                    if (!pids.ContainsKey(p))
                        pids.Add(p, new PidControllers());
                }
            }
            return c;
        }

        public void RegisterBehaviour(string moduleName, List<ModuleInfo> list) {
            foreach (Part p in system.vessel.parts) {
                if (p.Modules.Contains(moduleName)) {
                    ModuleInfo mi = list.Find(x => x.name == p.partInfo.name);
                    if (mi != null && !mi.behaviours.Contains(moduleName)) {
                        mi.behaviours.Add(moduleName);
                        if (moduleName == "LazorTriggerContinous") {
                            mi.triggerName = p.Modules["LazorTriggerContinous"].Fields.GetValue<string>("triggerName");
                        }
                        if (moduleName == "LazorWarhead") {
                            mi.proximitydetonate_range = p.Modules["LazorWarhead"].Fields.GetValue<float>("blastRadius") / 2; //adding it a lot but it dont matter
                            mi.proximitydetonate_maxrange = p.Modules["LazorWarhead"].Fields.GetValue<float>("blastRadius"); //adding it a lot but it dont matter
                        }
                    }
                }

            }
        }

        //mainly to check whether modulesCounterparts contains a part in its value list of lists
        private bool DictionaryContains(Dictionary<int, List<Part>>.ValueCollection x, Part p) {
            foreach (List<Part> lp in x)
                foreach (Part px in lp)
                    if (px == p)
                        return true;
            return false;
        }

		#endregion

    }
}
