/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/
using System;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;

namespace Romfarer
{
	/// <summary>
	/// Lazor system controller.
	/// Shared by all LazorSystems on a Vessel
	/// object server for lazor systems, has references to each system and handles switching between active systems.
	/// </summary>
	class LazorSystemController {

		public Vessel vessel;
		private LinkedList<LazorSystem> lazorSystems = new LinkedList<LazorSystem>(); //list of lazor systems on a vessel
		private HashSet<LazorSystem> lazorSystemsSet = new HashSet<LazorSystem>(); //set of lazor systems on a vessel
		public LazorSystem ActiveLazorSystem { get { return lazorSystems.First.Value; } }

		public LazorSystemController(Vessel vessel, LazorSystem system) {
			RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Creating LSC : " + system, true);
			this.vessel = vessel;
			AddLast(system);
			foreach(Part p in vessel.parts)
				if(p is LazorSystem && p != system)
					AddLast(p as LazorSystem);
			LazorSystemControllerGlobal.Register_LazorSystemController(this);
		}

		public void AddLast(LazorSystem system){
			RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Adding : " + system, true);
			lazorSystems.AddLast(system);
			lazorSystemsSet.Add(system);
			system.lazorSystemController = this;
			system.sysLock_WaitDisconnect.Unlock();
			system.sysLock_WaitCouple.Unlock();
		}

		public void Remove(LazorSystem system) {
			if(lazorSystems.Remove(system) && lazorSystemsSet.Remove(system))
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Removed : " + system, true);
			else
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Removed FAILED: " + system, true);
			if(lazorSystems.Count == 0) {
				LazorSystemControllerGlobal.Remove_LazorSystemController(this);
			}
		}

		public bool IsActiveLazorSystem(LazorSystem s) {
			return FlightGlobals.ActiveVessel == s.vessel && lazorSystems.Count != 0 && lazorSystems.First.Value == s;
		}

		public bool IsOnActiveVessel(LazorSystem s) {
			return FlightGlobals.ActiveVessel == s.vessel && lazorSystemsSet.Contains(s);
		}

		public void SetActive(LazorSystem s) {
			if(ActiveLazorSystem != null)
				ActiveLazorSystem.TurnOffRemoteControl();
			s.TurnOffRemoteControl();
			lazorSystems.Remove(s);
			lazorSystems.AddFirst(s);
		}

		public LinkedList<LazorSystem>.Enumerator GetActiveSystemEnumerator() {
			return lazorSystems.GetEnumerator();
		}

		public void EnableSystemLock(Funx<LazorSystem, SystemLock> sysLock){
			foreach(LazorSystem sys in lazorSystems) {
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Enable " + sysLock(sys) + " : " + sys, true);
				sysLock(sys).Lock();
			}
		}
 
		public void DisableSystemLock(Funx<LazorSystem, SystemLock> sysLock, Action<LazorSystem> onUnlock) {
			for(int i = 0; i < lazorSystems.Count; i++) {
				LazorSystem sys = lazorSystems.ElementAt(i);
				onUnlock(sys);
				if(sys.vessel != vessel){
					RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Disable " + sysLock(sys) + " : sys on other vessel " + sys, true);
					Remove(sys);
					if(sysLock(sys))
						sys.lazorSystemController = new LazorSystemController(sys.vessel, sys);
				}else{
					RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LSC["+vessel.vesselName+"] : Disable " + sysLock(sys) + " : " + sys, true);
					sysLock(sys).Unlock();
				}
			}
		}
	}












	static class LazorSystemControllerGlobal {

		static public Vessel LastVesselWithLazor;
		static private Dictionary<Vessel, LazorSystemController> lazorSystemVessels = new Dictionary<Vessel, LazorSystemController>();

		static public LazorSystem ActiveLazorSystem { 
			get { 
				if(FlightGlobals.ActiveVessel != null && lazorSystemVessels.ContainsKey(FlightGlobals.ActiveVessel))
					return lazorSystemVessels[FlightGlobals.ActiveVessel].ActiveLazorSystem;
				else
					return null;
			} 
		}

		public static Component ActiveLazorSystemTarget {
			get {
				if(ActiveLazorSystem == null)
					return null;
				return ActiveLazorSystem.Target;
			}
		}

		static public void Register_LazorSystemController(LazorSystemController controller) {
			if(lazorSystemVessels.ContainsKey(controller.vessel)) {
				RuntimeLogic.Instance.Log("(f"+Time.frameCount+")LazorSystemControllerGlobal : ERROR controllers vessel already setup '"+controller.vessel.vesselName, true);
			} else {
				RuntimeLogic.Instance.Log("(f"+Time.frameCount+")LazorSystemControllerGlobal : Adding controller '"+controller.vessel.vesselName, true);
				lazorSystemVessels.Add(controller.vessel, controller);
			}
		}

		static public void Remove_LazorSystemController(LazorSystemController controller) {
			if(lazorSystemVessels.ContainsKey(controller.vessel)) {
				RuntimeLogic.Instance.Log("(f"+Time.frameCount+")LazorSystemControllerGlobal : Removing controller '"+controller.vessel.vesselName, true);
				lazorSystemVessels.Remove(controller.vessel);
			} else {
				RuntimeLogic.Instance.Log("(f"+Time.frameCount+")LazorSystemControllerGlobal : ERROR controller don't exist '"+controller.vessel.vesselName, true);
			}
		}
		
		static public bool ActiveVesselHasLazorSystem() {
			return FlightGlobals.ActiveVessel && lazorSystemVessels.ContainsKey(FlightGlobals.ActiveVessel);
		}

		static public bool ActiveVesselHasLazorSystemAndItIsOn() {
			return ActiveVesselHasLazorSystem() && lazorSystemVessels[FlightGlobals.ActiveVessel].ActiveLazorSystem.lazorOn;
		}

		static public void MergeLazorControllers(Vessel from, Vessel to) {
			RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")LazorSystemControllerGlobal : Set Active from " + (from != null ? from.vesselName : "nothing") + " to " + (to != null ? to.vesselName : "nothing"), true);
			if(lazorSystemVessels.ContainsKey(to)) {
				if(from != null) {
					if(lazorSystemVessels.ContainsKey(from)) { //merge from -> to
						using(LinkedList<LazorSystem>.Enumerator systems = lazorSystemVessels[from].GetActiveSystemEnumerator())
							while(systems.MoveNext())
								lazorSystemVessels[to].AddLast(systems.Current);
						lazorSystemVessels.Remove(from);
					}
					lazorSystemVessels [to].EnableSystemLock ((sys)=>{return sys.sysLock_WaitCouple;});
				}
			} else
				SetupLazorSystemControllerOnVessel(to);
		}

		static public void EnableSystemLock(Vessel vessel, Funx<LazorSystem, SystemLock> sysLock){
			LazorSystemController lsc = GetController(vessel);
			if(lsc != null){
				lsc.EnableSystemLock ((sys)=>{return sys.sysLock_WaitDisconnect;});
			}
		}

		//assumes 'vessel' has no controller set up already
		static private void SetupLazorSystemControllerOnVessel(Vessel vessel) {
			foreach(Part p in vessel.parts) {
				if(p is LazorSystem) {
					(p as LazorSystem).lazorSystemController = new LazorSystemController(vessel, p as LazorSystem);
					return;
				}
			}
		}

		static public LazorSystemController GetController(Vessel vessel){
			if(vessel != null && lazorSystemVessels.ContainsKey(vessel))
				return lazorSystemVessels[vessel];
			return null;
		}

	}






	public class SystemLock{
		protected bool locked;
		public string name;

		public SystemLock(string name) {
			this.name = name;
		}

		virtual public void Lock() {
			locked = true;
		}

		virtual public void Unlock() {
			locked = false;
		}

		virtual public bool Locked { get {  return locked; } }

		static public implicit operator bool(SystemLock s) {
			return s.Locked;
		}
		static public implicit operator string(SystemLock s) {
			return s.name;
		}
	}

	public class SystemLockTimed : SystemLock{
		private int frame = int.MaxValue;

		public SystemLockTimed(string name) : base (name) {}
		
		override public void Lock() {
			locked = true;
			frame = Time.frameCount;
		}
		
		override public void Unlock() {
			locked = false;
			frame = int.MaxValue;
		}
		
		override public bool Locked { get {  return Locked_bool && Locked_frame; } }
		
		public bool Locked_bool  { get { return locked; } }
		
		public bool Locked_frame { get { return frame >= Time.frameCount; } }

	}


}

