﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {

	interface IPid{
		float Update(float v, float dt);
	}

    class PID : IPid{

	    private float	m_ControlValue = 0;
	    private float	kp;
	    private float	ki;
	    private float	kd;
        public float KP { get { return kp; } }
        public float KI { get { return ki; } }
        public float KD { get { return kd; } }
	    private float	e0, e1, e2 = 0;
        private float i = 0;


	    public PID( float kp, float ki, float kd ){
            this.kp = kp;
            this.ki = ki;
            this.kd = kd;
        }

        public void Reinitialize(float kp, float ki, float kd) {
            this.kp = kp;
            this.ki = ki;
            this.kd = kd;
        }
 
        //sv = target value.
        //pv = current/measured value.
        //dt = time elapsed since last update.
        //public float Update( float sv, float pv, float dt ){
        public float Update( float v, float dt ){
            if ( dt <= 0 )
		        return -1;

	        e1 = e0;
            e0 = v;
	        float e01 = e0 - e1;
	        float e12 = e1 - e2;
	        float p	= kp * e0;

            i += e0 * dt;
	        float d	= kd * (e0 - e1) / dt;

            return p + ki * i + d;
 
        }

        public static implicit operator float(PID v) {
            return v.m_ControlValue;
        }


    }




	class PIDclamp  : IPid{

        private float output = 0;
        private float kp, ki, kd;
        public float kpT, kiT, kdT; //temp settings
        public float KP { get { return kp; } }
        public float KI { get { return ki; } }
        public float KD { get { return kd; } }
        private float clamp;
        private float integral;
        private float previous_integral;
        private float previous_error;
        private float previous_output;
		public string name;


		public PIDclamp(string name, float kp, float ki, float kd, float clamp = 1) {
            this.kp = kp;
            this.ki = ki;
            this.kd = kd;
            kpT = kp;
            kiT = ki;
            kdT = kd;
            this.clamp = clamp;
			this.name = name;

        }

        public void Reinitialize(float kp, float ki, float kd) {
			ReinitializePIDsOnly(kp, ki, kd);
            integral = 0;
            previous_error = 0;
        }

		public void ReinitializePIDsOnly(float kp, float ki, float kd) {
			this.kp = kp;
			this.ki = ki;
			this.kd = kd;
			kpT = kp;
			kiT = ki;
			kdT = kd;
		}

		public void Reinitialize(string name, float kp, float ki, float kd) {
			this.name = name;
			Reinitialize(kp, ki, kd);
		}

		public void Clamp(float clamp) {
			this.clamp = clamp;
		}

		/*public void Reinitialize(PIDclamp persist) {
			this.kp = persist.kp;
			this.ki = ki;
			this.kd = kd;
			kpT = kp;
			kiT = ki;
			kdT = kd;
			this.clamp = clamp;
			this.name = name;
		}*/

		float pre = 0;
        public float UpdatePD(float error, float dt, float clampI) {
            previous_integral = integral;
            previous_output = output;

			float derivative = (error - previous_error) / dt;

			float m = Math.Sign(error == 0 ? 1 : error);
			float mP = Mathf.Clamp(error * error + 1, 1f, 1000);
			float mD = Mathf.Clamp(Math.Abs(derivative * derivative) + 0.5f, 0.5f, 1000);

			if(Math.Abs(error) < clampI)
				integral += Mathf.Clamp(0.005f / (m * mD), -0.1f, 0.1f);
            	//integral += error * dt;
			integral = Mathf.Clamp(integral, -clampI * 4, clampI * 4);
            
            previous_error = error;
            output = (kp * error) + (ki * integral) + (kd * derivative);

			//MonoBehaviour.print((Time.time - pre)+" $pid i " + integral + " m " + m+ " mP " +mP+ " mD " +mD);
			pre = Time.time;
			//MonoBehaviour.print("$pid e "+error + " i " + integral + " d' " + derivative+ " dt " +dt);

            if (output > -clamp && output < clamp)
                return output;
            else {
                integral = previous_integral;
                output = Mathf.Clamp(output, -clamp, clamp);
                return output;
            }
        }

		public float Update(float error, float dt) {

            previous_integral = integral;
            previous_output = output;
            integral += error * dt;
            float derivative = (error - previous_error) / dt;
            previous_error = error;
            output = (kp * error) + (ki * integral) + (kd * derivative);

			//MonoBehaviour.print("$pid e "+error + " i " + integral + " d' " + derivative+ " dt " +dt);

            if (output > -clamp && output < clamp)
                return output;
            else {
                integral = previous_integral;
                output = Mathf.Clamp(output, -clamp, clamp);
                return output;
            }
        }

        public static implicit operator float(PIDclamp v) {
            return v.output;
        }
		
        public void DebugGui(string title) {
            GUILayout.Label(title);
            GUILayout.BeginHorizontal(Util.Space_Compact);
            kpT = float.Parse(GUILayout.TextField(kpT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
            kiT = float.Parse(GUILayout.TextField(kiT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
            kdT = float.Parse(GUILayout.TextField(kdT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
            GUILayout.EndHorizontal();
            if (GUILayout.Button("Reinitialize"))
                Reinitialize(kpT, kiT, kdT);
        }

		public void DebugGui2(string title, out float p, out float i, out float d) {
			GUILayout.Label(title);
			GUILayout.BeginHorizontal(Util.Space_Compact);
			p = kpT = float.Parse(GUILayout.TextField(kpT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
			i = kiT = float.Parse(GUILayout.TextField(kiT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
			d = kdT = float.Parse(GUILayout.TextField(kdT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
			GUILayout.EndHorizontal();
		}

		public static void DebugGuiPanel(params PIDclamp[] pids) {
			GUILayout.BeginVertical();
			foreach(PIDclamp p in pids) {
				GUILayout.Label(p.name);
				GUILayout.BeginHorizontal(Util.Space_Compact);
				p.kpT = float.Parse(GUILayout.TextField(p.kpT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
				p.kiT = float.Parse(GUILayout.TextField(p.kiT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
				p.kdT = float.Parse(GUILayout.TextField(p.kdT.ToString("f6"), GUILayout.Width(63)).Replace("\n", "").Replace("\r", ""));
				GUILayout.EndHorizontal();
			}
			GUILayout.Space(5);
			GUILayout.BeginHorizontal();
			if(GUILayout.Button("Reinitialize"))
				foreach(PIDclamp p in pids)
					p.Reinitialize(p.kpT, p.kiT, p.kdT);
			if(GUILayout.Button("Save"))
				LazorGlobals.SaveConfig();
			GUILayout.EndHorizontal();
			GUILayout.EndVertical();
		}

		override public string ToString(){
			return name + "#" + kp.ToString() + "#" + ki.ToString() + "#" + kd.ToString() + "#" + clamp.ToString();
		}

		public static PIDclamp Parse(string s) {
			char delimeter = '#';
			string[] p = s.Split(delimeter);
			return new PIDclamp(p[0], float.Parse(p[1]), float.Parse(p[2]), float.Parse(p[3]), float.Parse(p[4]));
		}
    }
}
