﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    abstract class VesselIterator {
        public LazorSystem system;
        public Vessel vessel;
        public float mass = 0f;
        public Dictionary<string, Resource> resourceParts = new Dictionary<string, Resource>();

        internal class Resource {
            public List<Part> parts = new List<Part>();
            public double amount = 0;
            public double amountPre = 0;
            public double maxAmount = 0;
            public double capacity { //available space for this resource
                get { return maxAmount - amount; }
            }
            public UtilAverage flow = new UtilAverage(40); //positive for draining, negative for gaining
            public Resource() { }
            public void Clear() {
                amount = 0;
                maxAmount = 0;
            }
        }

        public VesselIterator(LazorSystem system, Vessel vessel) {
            this.system = system;
            this.vessel = vessel;
        }

		abstract public void Update(bool force);

        public static List<Part> GetEngines(Vessel vessel) {
            List<Part> list = new List<Part>();
            foreach (Part p in vessel.parts)
                foreach (PartModule m in p.Modules)
                    if (m is ModuleEngines)
                        list.Add(p);
            return list;
        }

        public double DrainResource(string resource, float amount) {
            if (vessel != FlightGlobals.ActiveVessel)
                Update(true);
            double returnAmount = 0;
            double amountToBeDrained = amount;
            if (vessel.isEVA) {
                if (amount < resourceParts[resource].amount) {
                    returnAmount += amount;
					KerbalEVA eva = (vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA);
					eva.part.Resources.Get(eva.propellantResourceId).amount -= amount;

					//(vessel.rootPart.Modules["KerbalEVA"] as KerbalEVA).Fuel -= amount;
                } else {
                    returnAmount += resourceParts[resource].amount;
					KerbalEVA eva = (vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA);
					eva.part.Resources.Get(eva.propellantResourceId).amount = 0;
					//(vessel.rootPart.Modules["KerbalEVA"] as KerbalEVA).Fuel = 0;
                }
                return returnAmount;
            } else {
                List<Part> list = new List<Part>();
                List<Part> list2 = new List<Part>(vessel.rootPart.children);
                list2.Add(vessel.rootPart);
                while (true) {
                    list = new List<Part>(list2);
                    //MonoBehaviour.print("children " + list.Count + " ret " + returnAmount.ToString("f2") + " remain " + amountToBeDrained.ToString("f2"));
                    if (list.Count > 0) {
                        List<Part> partsWithResource = list.FindAll(x => x.Resources.Contains(resource) && x.Resources[resource].amount > 0);
                        if (partsWithResource.Count > 0) {
                            double a = amountToBeDrained / partsWithResource.Count;
                            double remainder = 0; //left in this group
                            foreach (Part p in list) {
                                if (p.Resources.Contains(resource)) {
                                    PartResource ft = p.Resources[resource] as PartResource;
                                    //MonoBehaviour.print("tank " + " a " + a.ToString("f2") + " ft " + ft.amount.ToString("f2") + " r " + remainder);
                                    if (a < ft.amount) {
                                        remainder += Mathf.Clamp((float)(ft.amount - a), 0, (float)ft.amount);
                                        returnAmount += a;
                                        ft.amount -= a;
                                        amountToBeDrained -= a;
                                    } else {
                                        amountToBeDrained -= ft.amount;
                                        returnAmount += ft.amount;
                                        ft.amount = 0;
                                    }
                                }
                            }
                            if (amountToBeDrained < 1) {
                                return returnAmount;
                            } else {
                                if (remainder < 0.1d) {
                                    list2.Clear();
                                    foreach (Part p in list)
                                        list2.AddRange(p.children);
                                }
                            }
                        } else {
                            list2.Clear();
                            foreach (Part p in list)
                                list2.AddRange(p.children);
                        }
                    } else
                        return returnAmount;
                }
            }
        }

        public void AddResource(string resource, float amount) {
            if (vessel != FlightGlobals.ActiveVessel)
                Update(true);
            if (vessel.isEVA) {
				if (resourceParts [resource].amount + amount < 100) {
					//(vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA).Fuel += amount;
					KerbalEVA eva = (vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA);
					eva.part.Resources.Get(eva.propellantResourceId).amount += amount;
				} else {
					//(vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA).Fuel = 100;
					KerbalEVA eva = (vessel.rootPart.Modules ["KerbalEVA"] as KerbalEVA);
					eva.part.Resources.Get(eva.propellantResourceId).amount = eva.FuelCapacity;
				}
            } else {
                List<Part> list = new List<Part>();
                List<Part> list2 = new List<Part>(vessel.rootPart.children);
                list2.Add(vessel.rootPart);
                while (true) {
                    list = new List<Part>(list2);
                    if (list.Count > 0) {
                        float a = amount / list.FindAll(x => x.Resources.Contains(resource)).Count;
                        bool full = true;
                        foreach (Part p in list) {
                            if (p.Resources.Contains(resource)) {
                                PartResource ft = (PartResource)p.Resources[resource];
                                if (ft.amount + a < ft.maxAmount) {
                                    ft.amount += a;
                                    full = false;
                                } else
                                    ft.amount = ft.maxAmount;
                            }
                        }
                        if (full) {
                            list2.Clear();
                            foreach (Part p in list)
                                if (p.children.Count > 0)
                                    list2.AddRange(p.children);
                        } else
                            break;
                    } else
                        break;
                }
            }
        }

    }





    /// <summary>
    /// VesselIterator for your own vessel
    /// </summary>
    class VesselIteratorOwn : VesselIterator{

        public delegate void VesselOperation(Part p);
        private List<VesselOperation> tasks = new List<VesselOperation>();
        private Dictionary<uint, float> capacitiesFuel = new Dictionary<uint, float>();
        private Dictionary<uint, float> capacitiesRCS = new Dictionary<uint, float>(); 
        public int stage = int.MaxValue;
        public List<Part> resourcesInParts = new List<Part>(); //parts with resources
        public List<String> resourceTypes = new List<string>();
        //Parameters
        public LazorParameters lp;
        public List<LazorParameters.ResourceWrapper> resourceWrappers = new List<LazorParameters.ResourceWrapper>();
        //general
        public LazorParameters.ParameterWrapper param_MissionTime;
        public LazorParameters.ParameterWrapper param_Crew;
        public LazorParameters.ParameterWrapper param_Situation;
        public LazorParameters.ParameterWrapper param_SoI;
        //surface
        public LazorParameters.ParameterWrapper param_AltitudeFT;
        public LazorParameters.ParameterWrapper param_SurfaceSpeed;
        public LazorParameters.ParameterWrapper param_VerticalSpeed;
        public LazorParameters.ParameterWrapper param_Latitude;
        //public LazorParameters.ParameterWrapper param_Longitude;
        //orbit
        public LazorParameters.ParameterWrapper param_Apoapsis;
        public LazorParameters.ParameterWrapper param_Acceleration;
        public LazorParameters.ParameterWrapper param_OrbitSpeed;
        public LazorParameters.ParameterWrapper param_Altitude;
        public LazorParameters.ParameterWrapper param_TimeToApoapsis;
        public LazorParameters.ParameterWrapper param_Periapsis;
        public LazorParameters.ParameterWrapper param_TimeToPeriapsis;
        public LazorParameters.ParameterWrapper param_Period;
        public LazorParameters.ParameterWrapper param_Eccentricity;
        public LazorParameters.ParameterWrapper param_Inclination;
        public LazorParameters.ParameterWrapper param_MeanAnomaly;
        public LazorParameters.ParameterWrapper param_EccentricAnomaly;
        public LazorParameters.ParameterWrapper param_TrueAnomaly;
        //physical
        public LazorParameters.ParameterWrapper param_CurrentStage;
        public LazorParameters.ParameterWrapper param_AtmosphericDensity;
        public LazorParameters.ParameterWrapper param_Pressure;
        public LazorParameters.ParameterWrapper param_DynamicPressure;
        public LazorParameters.ParameterWrapper param_GeeForce;
        public LazorParameters.ParameterWrapper param_Gravity;
        public LazorParameters.ParameterWrapper param_Temperature;
        public LazorParameters.ParameterWrapper param_Mass;
        public bool hasFuel;
        
        public void X(LazorParameters.ParameterWrapper x){
            x.X();
        }

        public void Y() {
            foreach (LazorParameters.ResourceWrapper r in resourceWrappers)
                r.X();
        }

        public VesselIteratorOwn(LazorSystem system, Vessel vessel) : base(system, vessel){
            stage = vessel.currentStage;
            lp = new LazorParameters(system, this, system.vessel.mainBody);
            foreach (string s in resourceTypes)
                resourceWrappers.Add(new LazorParameters.ResourceWrapper(lp.ResourceItem, false, s));
            //general
            param_MissionTime = new LazorParameters.ParameterWrapper(lp.MissionTime, false);
            param_Crew = new LazorParameters.ParameterWrapper(lp.Crew, false);
            param_Situation = new LazorParameters.ParameterWrapper(lp.Situation, false);
            param_SoI = new LazorParameters.ParameterWrapper(lp.SoI, false);
            //surface
            param_AltitudeFT = new LazorParameters.ParameterWrapper(lp.AltitudeFT, false);
            param_SurfaceSpeed = new LazorParameters.ParameterWrapper(lp.SurfaceSpeed, false);
            param_VerticalSpeed = new LazorParameters.ParameterWrapper(lp.VerticalSpeed, false);
            param_Latitude = new LazorParameters.ParameterWrapper(lp.Latitude, false);
            //param_Longitude = new LazorParameters.ParameterWrapper(lp.Longitude, false);
            //orbit
            param_Acceleration = new LazorParameters.ParameterWrapper(lp.Acceleration, false);
            param_OrbitSpeed = new LazorParameters.ParameterWrapper(lp.OrbitSpeed, false);
            param_Altitude = new LazorParameters.ParameterWrapper(lp.Altitude, false);
            param_Apoapsis = new LazorParameters.ParameterWrapper(lp.Apoapsis, false);
            param_TimeToApoapsis = new LazorParameters.ParameterWrapper(lp.TimeToApoapsis, false);
            param_Periapsis = new LazorParameters.ParameterWrapper(lp.Periapsis, false);
            param_TimeToPeriapsis = new LazorParameters.ParameterWrapper(lp.TimeToPeriapsis, false);
            param_Period = new LazorParameters.ParameterWrapper(lp.Period, false);
            param_Eccentricity = new LazorParameters.ParameterWrapper(lp.Eccentricity, false);
            param_Inclination = new LazorParameters.ParameterWrapper(lp.Inclination, false);
            param_MeanAnomaly = new LazorParameters.ParameterWrapper(lp.MeanAnomaly, false);
            param_EccentricAnomaly = new LazorParameters.ParameterWrapper(lp.EccentricAnomaly, false);
            param_TrueAnomaly = new LazorParameters.ParameterWrapper(lp.TrueAnomaly, false);
            //physical
            param_CurrentStage = new LazorParameters.ParameterWrapper(lp.CurrentStage, false);
            param_AtmosphericDensity = new LazorParameters.ParameterWrapper(lp.AtmosphericDensity, false);
            param_Pressure = new LazorParameters.ParameterWrapper(lp.Pressure, false);
            param_DynamicPressure = new LazorParameters.ParameterWrapper(lp.DynamicPressure, false);
            param_GeeForce = new LazorParameters.ParameterWrapper(lp.GeeForce, false);
            param_Gravity = new LazorParameters.ParameterWrapper(lp.Gravity, false);
            param_Temperature = new LazorParameters.ParameterWrapper(lp.Temperature, false);
            param_Mass = new LazorParameters.ParameterWrapper(lp.Mass, false);
            ResetResourceLists();
        }

        public void ResetResourceLists() {
            resourceParts.Clear();
            resourceTypes.Clear();
            resourcesInParts.Clear();

            //add the parts that has the resource to list
            foreach (Part p in vessel.parts) {
                if (!p.Modules.Contains("ModuleEngines")) { //engines has ElectricCharge resource, but it can not store electricity.
                    if (p.Resources.list.Count > 0)
                        resourcesInParts.Add(p);
                    foreach (PartResource r in p.Resources.list) {
                        if (!resourceParts.ContainsKey(r.resourceName)) {
                            resourceParts.Add(r.resourceName, new Resource());
                            resourceTypes.Add(r.resourceName);
                        }
                        resourceParts[r.resourceName].parts.Add(p);
                    }
                }
            }
            resourceWrappers.Clear();
            foreach (string s in resourceTypes)
                resourceWrappers.Add(new LazorParameters.ResourceWrapper(lp.ResourceItem, false, s));
            hasFuel = resourceTypes.Contains("LiquidFuel");
        }

        public void AddTask(VesselOperation p) {
            tasks.Add(p);
        }

        public void RemoveTask(VesselOperation p) {
            tasks.Remove(p);
        }

        //runs through all parts on the vessel and gather data;
        override public void Update(bool force = false) {
            mass = 0;
            foreach (Part p in vessel.parts) {
                mass += p.mass;
                foreach (VesselOperation vo in tasks)
                    vo(p);
            }
            if (vessel == FlightGlobals.ActiveVessel || force) {
                foreach (Resource resource in resourceParts.Values)
                    resource.Clear();
                foreach (Part p in resourcesInParts) { //sum totals
                    if (p.localRoot != vessel.rootPart) { //part must have been broken off
                        MonoBehaviour.print("VesselIterator ... part not attached to vessel");
                        ResetResourceLists();
                        return;
                    }
                    foreach (PartResource r in p.Resources.list) {
                        Resource resource = resourceParts[r.resourceName];
                        resource.amount += r.amount;
                        resource.maxAmount += r.maxAmount;
                    }
                }
                foreach (Resource resource in resourceParts.Values) { //calculate ratio from totals
                    resource.flow.Average((resource.amountPre - resource.amount) / Time.deltaTime);
                    resource.amountPre = resource.amount;
                }
            }
            //events
            if (stage > vessel.currentStage) {
                system.localEvents.RegisterStageChange(this); //fire stage change event
                stage = vessel.currentStage;
				ResetResourceLists(); //Hardcoded event
            }
        }


        
        


    }





































    /// <summary>
    /// Vessel iterator for the target vessel
    /// </summary>
    class VesselIteratorTarget : VesselIterator {

        public delegate void VesselOperation(Part p);
        private List<VesselOperation> tasks = new List<VesselOperation>();
        private Dictionary<uint, float> capacitiesFuel = new Dictionary<uint, float>();
        private Dictionary<uint, float> capacitiesRCS = new Dictionary<uint, float>();
        public int stage = int.MaxValue;
        public List<Part> resourcesParts = new List<Part>(); //parts with resources
        public List<String> resourceTypes = new List<string>();


        public LazorParameters lp;
        public LazorParameters.ParameterWrapper param_Distance;
        public LazorParameters.ParameterWrapper param_MissionTime;
        public LazorParameters.ParameterWrapper param_Crew;
        public LazorParameters.ParameterWrapper param_Situation;
        public LazorParameters.ParameterWrapper param_SoI;
        public LazorParameters.ParameterWrapper param_Altitude;
        public LazorParameters.ParameterWrapper param_OrbitSpeed;
        public LazorParameters.ParameterWrapper param_CurrentStage;
        public LazorParameters.ParameterWrapper param_Pressure;
        public LazorParameters.ParameterWrapper param_Apoapsis;
        public LazorParameters.ParameterWrapper param_TimeToApoapsis;
        public LazorParameters.ParameterWrapper param_Periapsis;
        public LazorParameters.ParameterWrapper param_TimeToPeriapsis;
        public LazorParameters.ParameterWrapper param_Period;
        public LazorParameters.ParameterWrapper param_Eccentricity;
        public LazorParameters.ParameterWrapper param_Inclination;
        //CelestialBody
        public LazorParameters.ParameterWrapper param_CB_Distance;
        public LazorParameters.ParameterWrapper param_CB_Radius;
        public LazorParameters.ParameterWrapper param_CB_Mass;
        public LazorParameters.ParameterWrapper param_CB_GravParam;
        public LazorParameters.ParameterWrapper param_CB_ASLGee;
        public LazorParameters.ParameterWrapper param_CB_HasAtmosphere;
        public LazorParameters.ParameterWrapper param_CB_AtmosAlt;
        public LazorParameters.ParameterWrapper param_CB_ASLPressure;
        public LazorParameters.ParameterWrapper param_CB_ASLDensity;
        public LazorParameters.ParameterWrapper param_CB_ASLTemperature;
        public LazorParameters.ParameterWrapper param_CB_HasOxygen;
        public LazorParameters.ParameterWrapper param_CB_HasOcean;
        public LazorParameters.ParameterWrapper param_CB_IsTidallyLocked;
        public LazorParameters.ParameterWrapper param_CB_IsRotating;
        public LazorParameters.ParameterWrapper param_CB_RotationPeriod;
        public LazorParameters.ParameterWrapper param_CB_IsRotatingInverse;
        public LazorParameters.ParameterWrapper param_CB_OrbitPeriod;
        public LazorParameters.ParameterWrapper param_CB_OrbitVelocity;
        public LazorParameters.ParameterWrapper param_CB_OrbitEccentricity;
        public LazorParameters.ParameterWrapper param_CB_OrbitInclination;
        public LazorParameters.ParameterWrapper param_CB_MeanAnomaly;
        public LazorParameters.ParameterWrapper param_CB_EccentricAnomaly;
        public LazorParameters.ParameterWrapper param_CB_TrueAnomaly;

        public void X(LazorParameters.ParameterWrapper x) {
            x.highlight = x.param(x.highlight);
        }

        public VesselIteratorTarget(LazorSystem system, Vessel vessel) : base(system, vessel) {
            stage = vessel.currentStage;
            OnVesselChange(); //Hardcoded event
            if (vessel != FlightGlobals.ActiveVessel) //when changing targets, get the amounts in target
                Update(true); //force the update
            lp = new LazorParameters(system, this, system.vessel.mainBody);
            param_Distance = new LazorParameters.ParameterWrapper(lp.Distance, false);
            param_MissionTime = new LazorParameters.ParameterWrapper(lp.MissionTime, false);
            param_Crew = new LazorParameters.ParameterWrapper(lp.Crew, false);
            param_Situation = new LazorParameters.ParameterWrapper(lp.Situation, false);
            param_SoI = new LazorParameters.ParameterWrapper(lp.SoI, false);
            param_Altitude = new LazorParameters.ParameterWrapper(lp.Altitude, false);
            param_OrbitSpeed = new LazorParameters.ParameterWrapper(lp.OrbitSpeed, false);
            param_CurrentStage = new LazorParameters.ParameterWrapper(lp.CurrentStage, false);
            param_Pressure = new LazorParameters.ParameterWrapper(lp.Pressure, false);
            param_Apoapsis = new LazorParameters.ParameterWrapper(lp.Apoapsis, false);
            param_TimeToApoapsis = new LazorParameters.ParameterWrapper(lp.TimeToApoapsis, false);
            param_Periapsis = new LazorParameters.ParameterWrapper(lp.Periapsis, false);
            param_TimeToPeriapsis = new LazorParameters.ParameterWrapper(lp.TimeToPeriapsis, false);
            param_Period = new LazorParameters.ParameterWrapper(lp.Period, false);
            param_Eccentricity = new LazorParameters.ParameterWrapper(lp.Eccentricity, false);
            param_Inclination = new LazorParameters.ParameterWrapper(lp.Inclination, false);
            //CelestialBody
            param_CB_Distance = new LazorParameters.ParameterWrapper(lp.CB_Distance, false);
            param_CB_Radius = new LazorParameters.ParameterWrapper(lp.CB_Radius, false);
            param_CB_Mass = new LazorParameters.ParameterWrapper(lp.CB_Mass, false);
            param_CB_GravParam = new LazorParameters.ParameterWrapper(lp.CB_GravParam, false);
            param_CB_ASLGee = new LazorParameters.ParameterWrapper(lp.CB_ASLGee, false);
            param_CB_HasAtmosphere = new LazorParameters.ParameterWrapper(lp.CB_HasAtmosphere, false);
            param_CB_AtmosAlt = new LazorParameters.ParameterWrapper(lp.CB_AtmosAlt, false);
            param_CB_ASLPressure = new LazorParameters.ParameterWrapper(lp.CB_ASLPressure, false);
            param_CB_ASLDensity = new LazorParameters.ParameterWrapper(lp.CB_ASLDensity, false);
            param_CB_ASLTemperature = new LazorParameters.ParameterWrapper(lp.CB_ASLTemperature, false);
            param_CB_HasOxygen = new LazorParameters.ParameterWrapper(lp.CB_HasOxygen, false);
            param_CB_HasOcean = new LazorParameters.ParameterWrapper(lp.CB_HasOcean, false);
            param_CB_IsTidallyLocked = new LazorParameters.ParameterWrapper(lp.CB_IsTidallyLocked, false);
            param_CB_IsRotating = new LazorParameters.ParameterWrapper(lp.CB_IsRotating, false);
            param_CB_RotationPeriod = new LazorParameters.ParameterWrapper(lp.CB_RotationPeriod, false);
            param_CB_IsRotatingInverse = new LazorParameters.ParameterWrapper(lp.CB_IsRotatingInverse, false);
            param_CB_OrbitPeriod = new LazorParameters.ParameterWrapper(lp.CB_OrbitPeriod, false);
            param_CB_OrbitVelocity = new LazorParameters.ParameterWrapper(lp.CB_OrbitVelocity, false);
            param_CB_OrbitEccentricity = new LazorParameters.ParameterWrapper(lp.CB_OrbitEccentricity, false);
            param_CB_OrbitInclination = new LazorParameters.ParameterWrapper(lp.CB_OrbitInclination, false);
            param_CB_MeanAnomaly = new LazorParameters.ParameterWrapper(lp.CB_MeanAnomaly, false);
            param_CB_EccentricAnomaly = new LazorParameters.ParameterWrapper(lp.CB_EccentricAnomaly, false);
            param_CB_TrueAnomaly = new LazorParameters.ParameterWrapper(lp.CB_TrueAnomaly, false);
        }

        private void OnVesselChange() {
            resourceParts.Clear();
            resourceTypes.Clear();
            resourcesParts.Clear();
            //add the parts that has the resource to list
            foreach (Part p in vessel.parts) {
                if (!p.Modules.Contains("ModuleEngines")) { //engines has ElectricCharge resource, but it can not store electricity.
                    if (p.Resources.list.Count > 0)
                        resourcesParts.Add(p);
                    foreach (PartResource r in p.Resources.list) {
                        if (!resourceParts.ContainsKey(r.resourceName)) {
                            resourceParts.Add(r.resourceName, new Resource());
                            resourceTypes.Add(r.resourceName);
                        }
                        resourceParts[r.resourceName].parts.Add(p);
                    }
                }
            }
        }

        public void AddTask(VesselOperation p) {
            tasks.Add(p);
        }
        public void RemoveTask(VesselOperation p) {
            tasks.Remove(p);
        }

        //runs through all parts on the vessel and gather data;
        override public void Update(bool force = false) {
            mass = 0;
            foreach (Part p in vessel.parts) {
                mass += p.mass;
                foreach (VesselOperation vo in tasks)
                    vo(p);
            }

            if (vessel == FlightGlobals.ActiveVessel || force) {
                foreach (Resource resource in resourceParts.Values) {
                    resource.Clear();
                }
                foreach (Part p in resourcesParts) { //sum totals
                    if (p.localRoot != vessel.rootPart) { //part must have been broken off
                        MonoBehaviour.print("VesselIterator ...  part not attached to vessel");
                        OnVesselChange();
                        return;
                    }
                    foreach (PartResource r in p.Resources.list) {
                        Resource resource = resourceParts[r.resourceName];
                        resource.amount += r.amount;
                        resource.maxAmount += r.maxAmount;
                    }
                }
                foreach (Resource resource in resourceParts.Values) { //calculate ratio from totals
                    resource.flow.Average((resource.amountPre - resource.amount) / Time.deltaTime);
                    resource.amountPre = resource.amount;
                }
            }
        }




    }
}
